from argparse import ArgumentParser, Namespace, SUPPRESS
import copy
from collections import defaultdict, Mapping
import importlib
import json
import re
from typing import Dict, Iterable, List, Optional, Set, Tuple, Union
import warnings
import yaml

from pampy import match, MatchError

from paranormal.params import *
from paranormal.units import unconvert_si_units


__all__ = ['Params', 'to_json_serializable_dict', 'from_json_serializable_dict', 'to_yaml_file',
           'from_yaml_file', 'create_parser_and_parse_args', 'to_argparse', 'from_parsed_args',
           'get_param_unit', 'merge_param_classes', 'append_params_attributes']


####################
# Params Interface #
####################


class Params(Mapping):
    """
    Base class for Params

    Supports:
    1. ** operation
    2 * operation (iterate through param names)

    Does Not Support:
    1. Non-direct inheritance (inheriting from a class that inherits from Params)

    Example Use Case:

    Inherit from this class and populate with Param classes from paranormal.params
    ```
    class A(Params):
        g = GeomspaceParam(help='test', default=[0, 100, 20])
        i = IntParam(help='my favorite int', required=True, choices=[1,2,3])
    ```
    """
    def __init__(self, **kwargs):
        cls = type(self)
        if cls not in Params.__subclasses__():
            raise NotImplementedError('Behavior is undefined for classes that do not directly '
                                      'inherit from Params.')
        _copy_nested_classes(self)
        for key, value in kwargs.items():
            if not key in cls.__dict__:
                raise KeyError(f'{cls.__name__} does not take {key} as an argument')
            setattr(self, key, value)
        _check_for_required_arguments(cls, kwargs)
        _ensure_properties_are_working(self)

    def __iter__(self):
        valid_keys = [k for k in self.__class__.__dict__.keys() if not k.startswith('_')]
        for key in valid_keys:
            yield key

    def __len__(self):
        return len(list(self.__iter__()))

    def __getitem__(self, item):
        return getattr(self, item)

    def __eq__(self, other) -> bool:
        return json.loads(json.dumps(to_json_serializable_dict(self))) == \
               json.loads(json.dumps(to_json_serializable_dict(other)))

    def si_set(self, param_name: str, value: Union[float, int, Iterable]) -> None:
        """
        Set the parameter from a value that's in SI units

        :param param_name: The parameter name to set
        :param value: The new value in SI units
        """
        param = type(self).__dict__.get(param_name)
        expanded_units = _expand_param_units(param)
        if expanded_units is not None:
            values = [unconvert_si_units(v, u) for v, u in zip(value, expanded_units)]
            setattr(self, param_name, values)
        else:
            unit = get_param_unit(type(self), param_name)
            setattr(self, param_name, unconvert_si_units(value, unit))

    def si_update(self, **kwargs) -> None:
        """
        Set parameters from multiple values in si units

        :param kwargs: parameter names and values to set
        """
        for k, v in kwargs.items():
            self.si_set(k, v)

    def update(self, **kwargs) -> None:
        """
        Set multiple parameters at once.

        :param kwargs: parameter names and values to set
        """
        for k, v in kwargs.items():
            setattr(self, k, v)


def get_param_unit(cls: type(Params), param_name: str) -> str:
    """
    Get the unit of a param
    """
    return getattr(cls.__dict__.get(param_name), 'unit', None)


def _check_for_required_arguments(cls: type(Params), kwargs: dict) -> None:
    """
    Make sure the arguments required by Params class or subclass exist in the kwargs dictionary
    used to instantiate the class
    """
    required_but_not_provided = []
    for k, v in cls.__dict__.items():
        if k.startswith('_') or not isinstance(v, BaseDescriptor):
            continue
        elif getattr(v, 'required', None) and not k in kwargs:
            required_but_not_provided.append(k)
    if required_but_not_provided != []:
        raise KeyError(f'{required_but_not_provided} are required arguments to instantiate '
                       f'{cls.__name__}')


def _ensure_properties_are_working(params: Params) -> None:
    """
    Evaluate the properties within a class to make sure they work based on the attributes
    """
    for k, v in params.items():
        if isinstance(v, property):
            getattr(params, k)


def _copy_nested_classes(params: Params):
    """
    Deepcopy and set any nested params classes to make sure users cannot mutate the nested classes.
    This function will mutate the passed in params class
    """
    for k, v in type(params).__dict__.items():
        if isinstance(v, Params):
            setattr(params, k, copy.deepcopy(v))


#################
# Serialization #
#################


def to_json_serializable_dict(params: Params,
                              *,
                              include_defaults: bool = True,
                              include_hidden_params: bool = False,
                              params_to_omit: Optional[Set[str]] = None) -> dict:

    """
    Convert Params class to a json serializable dictionary

    :param params: Params class to convert
    :param include_defaults: Whether or not to include the param attribute default values if the
        values haven't been set yet
    :param include_hidden_params: Whether or not to include params that start with _
    :param params_to_omit: Params to set to None when serializing the params class
    :return A dictionary that's json serializable
    """
    retval = {}
    for k, v in type(params).__dict__.items():
        if params_to_omit is not None and k in params_to_omit:
            continue
        if ((not k.startswith('_') or include_hidden_params) and
                (include_defaults or k in params.__dict__)):
            if isinstance(v, BaseDescriptor):
                retval[k] = v.to_json(params)
            elif isinstance(v, Params):
                nested_params = params.__dict__.get(k, v)
                _params_to_omit = _get_omitted_params(v)
                if _get_omitted_params(nested_params) - _params_to_omit != set():
                    warnings.warn('Params can only be marked as omitted inside the class definition'
                                  f' - ({_get_omitted_params(nested_params) - _params_to_omit})  '
                                  'will not be omitted.')
                retval[k] = to_json_serializable_dict(nested_params,
                                                      include_defaults=include_defaults,
                                                      include_hidden_params=include_hidden_params,
                                                      params_to_omit=_params_to_omit)
    retval['_type'] = params.__class__.__name__
    retval['_module'] = params.__class__.__module__
    return retval


def from_json_serializable_dict(dictionary: dict,
                                params_to_omit: Optional[Set[str]] = None) -> Params:
    """
    Convert from a json serializable dictionary to a Params class

    :param dictionary: a dictionary in the format returned by to_json_serializable_dict to convert
        back to a Params class
    :param params_to_omit: Params to set to None when de-serializing the dictionary
    :return The Params class
    """
    temp_dictionary = copy.deepcopy(dictionary)
    try:
        cls_type = temp_dictionary.pop('_type')
        module_name = temp_dictionary.pop('_module')
        module = importlib.import_module(module_name)
        cls = getattr(module, cls_type)
    except KeyError:
        raise ValueError('Dictionary Format Not Recognized!')
    # handle case with nested params
    unraveled_dictionary = {}
    for k, v in cls.__dict__.items():
        if isinstance(v, BaseDescriptor):
            if params_to_omit is not None and k in params_to_omit:
                unraveled_dictionary[k] = None
            elif k in temp_dictionary:
                unraveled_dictionary[k] = temp_dictionary[k]
        elif isinstance(v, Params):
            _params_to_omit = _get_omitted_params(v)
            unraveled_dictionary[k] = from_json_serializable_dict(temp_dictionary[k],
                                                                  params_to_omit=_params_to_omit)
    return cls(**unraveled_dictionary)


def to_yaml_file(params: Params,
                 filename: str,
                 *,
                 include_defaults: bool = False,
                 include_hidden_params: bool = False,
                 sort_keys: bool = False):
    """
    Dump to yaml
    """
    d = to_json_serializable_dict(params, include_defaults=include_defaults,
                                  include_hidden_params=include_hidden_params)

    with open(filename, 'w') as f:
        yaml.dump(d, stream=f, sort_keys=sort_keys)


def from_yaml_file(filename: str) -> Params:
    """
    Parse from yaml
    """
    with open(filename, 'r') as f:
        params = from_json_serializable_dict(yaml.full_load(f))
    return params


#############################
# Params Class Manipulation #
#############################


def _merge_positional_params(params_list: List[Tuple[str, BaseDescriptor]]
                             ) -> Tuple[List[str], Optional[BaseDescriptor]]:
    """
    Merge positional params into a single list param and return a list of param names that were
    merged

    :param params_list: List of params (name, Param) to search through for positional params. If
        multiple positional params are found, they will be merged into a single positional
        ListParam.
    :return The positional param names in the list, the merged ListParam
    """
    if not sum([getattr(p, 'positional', False) for (_, p) in params_list]) > 1:
        return [], None
    positionals = {k: v for (k, v) in params_list
                   if getattr(v, 'positional', False)}
    # Just parse all positionals as a ListParam - we'll match them back up later in
    # _parse_positional_arguments
    positional_param = ListParam(help=', '.join(v.help for v in positionals.values()
                                                if v.help != SUPPRESS),
                                 positional=True, required=True, nargs='+')
    return list(positionals.keys()), positional_param


def merge_param_classes(*cls_list,
                        merge_positional_params: bool = True) -> type(Params):
    """
    Merge multiple Params classes into a single merged params class and return the merged class.
    Note that this will not flatten the nested classes.

    :param cls_list: A list of Params subclasses or classes to merge into a single
        Params class
    :param merge_positional_params: Whether or not to merge the positional params in the classes
    """
    if len(cls_list) == 1:
        return cls_list[0]

    class MergedParams(Params):
        __doc__ = f'A Combination of {len(cls_list)} Params Classes:\n'

    append_params_attributes(MergedParams, *cls_list)
    for params_cls in cls_list:
        MergedParams.__doc__ += f'\n\t {params_cls.__name__} - {params_cls.__doc__}'

    # resolve positional arguments:
    if merge_positional_params:
        params_to_delete, positional_param = _merge_positional_params(
            [(k, v) for k, v in MergedParams.__dict__.items() if not k.startswith('_')])
        if positional_param is None and params_to_delete == []:
            return MergedParams
        setattr(MergedParams, 'positionals', positional_param)
        positional_param.__set_name__(MergedParams, 'positionals')
        for k in params_to_delete:
            delattr(MergedParams, k)

    return MergedParams


def append_params_attributes(cls: type(Params),
                             *other_cls_list,
                             do_not_copy: Optional[List[str]] = None,
                             override_dictionary: Optional[Dict] = None) -> None:
    """
    When building a Params subclass, copy parameters from other Params subclasses using this
    function (which will mutate the Params subclass but not the classes being copied from). If
    there are parameters you do not want to copy, specify them in the do_not_copy list, and if
    you'd like to override any defaults, specify them in the override_dictionary. Any merge
    conflicts will throw an error.

    :param cls: The params subclass to copy to
    :param other_cls_list: Any number or other params subclasses to copy from
    :param do_not_copy: A list of parameters not to copy
    :param override_dictionary: A dictionary specifying any overrides to parameters (help, default,
        etc.)


    Ex.

    ```python

    class A(Params):
        i = IntParam(help='int', default=1)

    class B(Params):
        f = FloatParam(help='float', default=2.0)
        z = IntParam(help='another int')


    append_params_attributes(A, B, do_not_copy=['z'],
                            override_dictionary={'f' : {'help': 'a better float', 'default': 3.0}})
    a = A()
    print(a.f)  # prints 3.0
    print(a.z)  # throws error
    ```
    """
    for other_cls in other_cls_list:
        for k, v in other_cls.__dict__.items():
            if not (k.startswith('_') or (do_not_copy is not None and k in do_not_copy)):
                if cls.__dict__.get(k, None) is not None:
                    raise ValueError(f'Unable to append params from classes {other_cls_list} due '
                                     f'to conflicting param: {k}')
                if isinstance(v, BaseDescriptor):
                    copied_v = copy.deepcopy(v)
                    copied_v.__set_name__(cls, k)
                elif isinstance(v, Params):
                    copied_v = copy.deepcopy(v)
                elif isinstance(v, property):
                    copied_v = v
                else:
                    copied_v = copy.deepcopy(v)
                if override_dictionary is not None and k in override_dictionary:
                    for _k, _v in override_dictionary[k].items():
                        setattr(copied_v, _k, _v)
                setattr(cls, k, copied_v)


####################
# Argument Parsing #
####################


def _get_param_type(param) -> type:
    """
    Get the expected type of a param
    """
    argtype = match(param,
                    LinspaceParam, lambda x: float,
                    SpanLinspaceParam, lambda x: float,
                    ArangeParam, lambda x: float,
                    SpanArangeParam, lambda x: float,
                    GeomspaceParam, lambda x: float,
                    ListParam, lambda x: param.subtype,
                    SetParam, lambda x: param.subtype,
                    EnumParam, lambda x: str,
                    IntParam, lambda x: int,
                    FloatParam, lambda x: float,
                    BoolParam, lambda x: None,
                    StringParam, lambda x: str,
                    BaseDescriptor, lambda x: type(getattr(param, 'default', None)))
    return argtype


def _convert_type_to_regex(argtype: type) -> str:
    """
    Source of truth for getting regex strings to match different types
    """
    regex_patterns = {int: r'\b[\+-]?(?<![\.\d])\d+(?!\.\d)\b',
                      float: r'[-\+]?(?:\d+(?<!\.)\.?(?!\.)\d*|\.?\d+)(?:[eE][-\+]?\d+)?',
                      str: r'\b.*\b'}
    return regex_patterns[argtype]


def _parse_positional_arguments(list_of_positionals: List[str],
                                positional_params: Dict[str, BaseDescriptor]) -> dict:
    """
    If params have been merged together, their positional arguments will have been merged as well
    into one single ListParam. This function's job is to pair the positionals with their respective
    unmerged param classes.

    :param list_of_positionals: A list of the positional arguments from the command line arguments
    :param positional_params: The positional parameters to match the positionals to
    :return: A dictionary that maps param name to its correct value in the positionals list
    """
    if list_of_positionals is None and positional_params != {}:
        raise ValueError('No positional arguments were returned from the parser')
    elif list_of_positionals is None and positional_params == {}:
        return {}
    # first loop through combs for params that have choices specified
    copied_pos_params = copy.deepcopy(positional_params)
    copied_pos_list = copy.deepcopy(list_of_positionals)
    correctly_matched_params = {}
    params_to_delete = []
    for name, param in copied_pos_params.items():
        choices = getattr(param, 'choices', None)
        if choices is not None:
            nargs = str(getattr(param, 'nargs', 1))
            # convert to raw string to handle escape characters
            matched = [p for p in copied_pos_list
                       if re.match(r'|'.join(fr'{c}' for c in choices), p)]
            if matched == []:
                raise ValueError(f'Could not match any of the choices: {choices} in '
                                 f'positionals: {list_of_positionals} for param: {name}')
            if not nargs in ['+', '*', '?']:
                # only take the first nargs arguments
                matched = matched[:int(nargs)]
            # convert to expected type
            if getattr(param, 'nargs', None) is not None:
                correctly_matched_params[name] = [_get_param_type(param)(p) for p in matched]
            else:
                correctly_matched_params[name] = _get_param_type(param)(matched[0])
            # remove matches
            for p in matched:
                copied_pos_list.remove(p)
            params_to_delete.append(name)
    # do garbage collection at the end to avoid mutating the list we're iterating through
    for n in params_to_delete:
        del copied_pos_params[n]

    # second loop through matches to type - for nargs +, *, and ?, the behavior is greedy
    params_to_delete = []
    # first sort unmatched params by least to most restrictive type (int, float, str)

    def _type_sorter(p) -> int:
        argtype = _get_param_type(p[1])
        order = defaultdict(lambda x: 4)
        order.update({int: 1, float: 2, str: 3})
        return order[argtype]

    for name, param in sorted(copied_pos_params.items(), key=_type_sorter):
        argtype = _get_param_type(param)
        pattern = _convert_type_to_regex(argtype)
        matched = [p for p in copied_pos_list if re.match(pattern, p)]
        if matched == []:
            raise ValueError(f'Unable to find {argtype} in {list_of_positionals} for param: {name}')
        nargs = str(getattr(param, 'nargs', 1))
        if not nargs in ['+', '*', '?']:
            matched = matched[:int(nargs)]
        # convert to expected type
        if getattr(param, 'nargs', None) is not None:
            correctly_matched_params[name] = [_get_param_type(param)(p) for p in matched]
        else:
            correctly_matched_params[name] = _get_param_type(param)(matched[0])
        # remove matches
        for p in matched:
            copied_pos_list.remove(p)
        params_to_delete.append(name)
    # do garbage collection at the end to avoid mutating the list we're iterating through
    for n in params_to_delete:
        del copied_pos_params[n]

    # No third Loop
    assert copied_pos_params == {} and copied_pos_list == [], 'Unable to parse positional arguments'
    return correctly_matched_params


def _add_param_to_parser(name: str, param: BaseDescriptor, parser: ArgumentParser) -> None:
    """
    Function to add a Param like IntParam, FloatParam, etc. called <name> to a parser

    :param name: The param name
    :param param: The param to add (IntParam, FloatParam, etc.)
    :param parser: The argument parser to add the param to.
    """
    argtype = _get_param_type(param)
    if argtype == type(None):
        raise NotImplementedError(f'Argparse type not implemented '
                                  f'for {param.__class__.__name__} and default not specifed')
    positional = getattr(param, 'positional', False)
    if getattr(param, 'prefix', '') != '' and not getattr(param, 'expand', False):
        raise ValueError(f'Failure with param {name}. Cannot add a prefix to a class without the'
                         f' expand kwarg set to True')
    argname = name if positional else '--' + name
    required = True if getattr(param, 'required', False) else None
    default = param.default if required is None else None
    unit = getattr(param, 'unit', None)

    # format metavar nicely for parameters that have expanded versions but aren't expanded
    # check to see if the parameter has expanded param names
    metavar = None
    try:
        expanded_param_names = _get_expanded_param_names(param)
    except MatchError:
        expanded_param_names = []
    if not getattr(param, 'expand', False) and expanded_param_names != []:
        metavar = tuple(x.upper() for x in expanded_param_names)

    # format help nicely if default is specified and suppress is not set
    help = param.help
    if not (positional or param.help == SUPPRESS):
        if unit is not None:
            help += f' [default: {default} {unit}]'
        else:
            help += f' [default: {default}]'
    if not required and positional:
        # TODO: use nargs='*' or nargs='?' to support not-required positional arguments
        raise ValueError('Not-required positional arguments are currently not supported')
    elif positional:
        # positional arguments are required by default, and argparse complains if you specify
        # required = True
        required = None
        default = None
    action = match(param,
                   BoolParam, lambda p: 'store_true' if not p.default else 'store_false',
                   BaseDescriptor, lambda p: None)
    nargs = getattr(param, 'nargs', None)
    assert not (action is not None and nargs is not None)
    choices = match(param,
                    EnumParam, lambda x: list(x.cls.__members__.keys()),
                    BaseDescriptor, lambda x: getattr(x, 'choices', None))
    kwargs = dict(action=action, nargs=nargs, default=default, metavar=metavar,
                  type=argtype, required=required, help=help, choices=choices)
    # we delete all kwargs that are None to avoid hitting annoying action class initializations
    # such as when action is store_true and 'nargs' is in kwargs
    for kw in list(kwargs.keys()):
        if kwargs[kw] is None:
            del kwargs[kw]
    parser.add_argument(argname, **kwargs)


def _get_omitted_params(params: Params) -> Set[str]:
    """
    Get a set of all params that have their value set to __hide__
    """
    return set(_k for _k in params.keys()
               if isinstance(params.__dict__.get(_k, None), str) and
               params.__dict__.get(_k, None) == '__hide__')


def _get_expanded_param_names(param: BaseDescriptor) -> List[str]:
    """
    Get bare expanded param names
    """
    return match(param,
                 SpanArangeParam, ['center', 'width', 'step'],
                 GeomspaceParam, ['start', 'stop', 'num'],
                 ArangeParam, ['start', 'stop', 'step'],
                 LinspaceParam, ['start', 'stop', 'num'],
                 SpanLinspaceParam, ['center', 'width', 'num'])


def _expand_param_name(param: BaseDescriptor) -> List[str]:
    """
    Get expanded param names

    :param param: The param to expand
    """
    if not getattr(param, 'expand', False):
        raise ValueError('Cannot expand param that does not have the expand kwarg')
    new_arg_names = _get_expanded_param_names(param)
    prefix = getattr(param, 'prefix', '')
    new_arg_names = [prefix + n for n in new_arg_names]
    return new_arg_names


def _expand_param_units(param: BaseDescriptor) -> List[Optional[str]]:
    """
    Get expanded param units

    :param param: The param to get units for
    :return: A list of the units (or None if no unit)
    """
    unit = getattr(param, 'unit', None)
    expanded_units = match(param,
                           GeomspaceParam,
                           lambda p: [unit, unit, None],
                           ArangeParam,
                           lambda p: [unit, unit, unit],
                           SpanArangeParam,
                           lambda p: [unit, unit, unit],
                           LinspaceParam,
                           lambda p: [unit, unit, None],
                           SpanLinspaceParam,
                           lambda p: [unit, unit, None],
                           BaseDescriptor, None)
    return expanded_units


def _expand_multi_arg_param(name: str, param: BaseDescriptor) -> Tuple[Tuple, Tuple, Tuple]:
    """
    Expand a parameter like GeomspaceParam or ArangeParam into seperate IntParams and FloatParams to
    parse as '--start X --stop X --num X' or '--start X --stop X --step X', etc.

    :param name: The param name
    :param param: The param to expand
    """
    new_arg_names = _expand_param_name(param)
    if getattr(param, 'positional', False):
        raise ValueError(f'Cannot expand positional {param.__class__.__name__} to {new_arg_names}')
    # expand the types
    expanded_types = match(param,
                           GeomspaceParam, [FloatParam, FloatParam, IntParam],
                           ArangeParam, [FloatParam, FloatParam, FloatParam],
                           SpanArangeParam, [FloatParam, FloatParam, FloatParam],
                           LinspaceParam, [FloatParam, FloatParam, IntParam],
                           SpanLinspaceParam, [FloatParam, FloatParam, IntParam])

    # expand the units
    expanded_units = _expand_param_units(param)

    # expand the defaults
    defaults = getattr(param, 'default', [None, None, None])
    if defaults is None:
        defaults = [None, None, None]

    # expand the choices
    choices = getattr(param, 'choices', None)
    expanded_choices = [[], [], []]
    if choices is not None:
        # transpose the choices for easy splitting but preserve type
        for c in choices:
            expanded_choices[0].append(c[0])
            expanded_choices[1].append(c[1])
            expanded_choices[2].append(c[2])

    # actually create the expanded params
    expanded_params = []
    for i, (default, argtype, u) in enumerate(zip(defaults, expanded_types, expanded_units)):
        new_param = argtype(help=param.help + ' (expanded into three arguments)',
                            default=default,
                            required=getattr(param,'required', False),
                            choices=expanded_choices[i] if choices is not None else None,
                            unit=u)
        expanded_params.append((new_arg_names[i], new_param))
    # we add the param we're expanding to the list as well to make sure the user does not pass it
    # Otherwise, it will get parsed silently, and the user will wonder what's happening
    copied_param = copy.deepcopy(param)
    setattr(copied_param, 'required', False)
    setattr(copied_param, 'help', SUPPRESS)
    setattr(copied_param, 'default', None)
    expanded_params.append((name, copied_param))
    return tuple(expanded_params)


def _extract_expanded_param(parsed_values: dict,
                            name: str,
                            param: BaseDescriptor,
                            prefix: Optional[str]) -> Optional[List]:
    """
    Convert [start, stop, num] or [start, stop, step], etc. from expanded form back into a list to
    be easily fed into the un-expanded param

    :param parsed_values: The parsed values from the command line as a dictionary (comes directly
        from the namespace)
    :param name: Original name of the param that was expanded (may have a prefix)
    :param param: The param that was expanded
    :param prefix: If the param is a nested param, there's a chance we had to append
        the enclosing class name as a prefix to deconflict arguments with the same name. See
        _create_param_name_prefix docstring for more info.
    """
    old_arg_names = [prefix + n if prefix is not None else n for n in _expand_param_name(param)]
    assert parsed_values.get(name) is None, f'param {name} was expanded! ' \
                                            f'Please provide {old_arg_names} instead'
    start_stop_x_list = [parsed_values[n] for n in old_arg_names]
    if all([x is None for x in start_stop_x_list]):
        return None
    return start_stop_x_list


def _create_param_name_prefix(enclosing_param_name: Optional[str],
                              prefix_dictionary: Optional[Dict]) -> str:
    """
    Create a param name variant to de-conflict conflicting params

    :param enclosing_param_name: The name of the enclosing class's parameter
    :param prefix_dictionary: If the user sets __nested_prefixes__ in their Params class
        declaration, they can manually override any prefix for a nested class or specify not to use
        one by specifying None.

    Ex.
    ```
    class A(Params):
        x = LinspaceParam(expand=True, ...)

    class B(Params):
        x = LinspaceParam(expand=True, ...)

    class C(Params):
        a = A()
        b = B()
    ```
    will result in a_start, a_stop, a_num, b_start, b_stop, and b_num as command line parameters
    because there's a conflict with x in both A and B.

    The enclosing param names for x are "a" and "b" in this example.

    Ex 2.
    ```
    class A(Params):
        x = LinspaceParam(expand=True, ...)

    class B(Params):
        x = LinspaceParam(expand=True, ...)

    class C(Params):
        __nested_prefixes__ = {'a': 'ayy', 'b': None}
        a = A()
        b = B()
    ```
    will result in ayy_start, ayy_stop, ayy_num, start, stop, and num as command line
    parameters.
    """
    if prefix_dictionary is not None and enclosing_param_name is not None:
        prefix = prefix_dictionary.get(enclosing_param_name, enclosing_param_name)
        return prefix + '_' if prefix is not None else ''
    return enclosing_param_name + '_' if enclosing_param_name is not None else ''


def _flatten_cls_params(cls: type(Params),
                        use_prefix: bool = True,
                        params_to_omit: Optional[Set[str]] = None,
                        defaults_to_overwrite: Optional[Dict] = None) -> Dict:
    """
    Extract params from a Params class - Behavior is as follows:

    1. Params with names starting in _ or with the hide attribute set to True will be ignored
    2. Properties and params in the params_to_omit set will be ignored
    3. If the class contains nested params classes, those will be flattened. Any nested class
        param set to __hide__ will be omitted
    4. If use_prefix is True, a prefix (the enclosing param name) will be prepended to any nested
        class parameter names no matter what.
    5. If there are any name conflicts during flattening, those will be resolved by prepending the
        enclosing param names as prefixes if use_prefix is True.
        See _create_param_name_prefix docstring for more details
    """
    already_flat_params = {}
    for name, param in vars(cls).items():
        # ignore params that start with _ or are in params_to_omit or have the hide attribute = True
        if (name.startswith('_') or
                (params_to_omit is not None and name in params_to_omit) or
                getattr(param, 'hide', False)):
            continue
        # if we have an actual param
        elif isinstance(param, BaseDescriptor):
            # overwrite the parameter default value if specified
            if defaults_to_overwrite is not None and name in defaults_to_overwrite:
                param = copy.deepcopy(param)  # need to deepcopy to avoid modifying the class
                setattr(param, 'default', defaults_to_overwrite[name])
            # if the param is supposed to be expanded, then expand it
            if getattr(param, 'expand', False):
                expanded = _expand_multi_arg_param(name, param)
                if any(n in already_flat_params for (n, _) in expanded):
                    raise ValueError('Please provide prefixes for the expanded params - conflict '
                                     f'with param: {name}')
                already_flat_params.update(dict(expanded))
            else:
                already_flat_params[name] = param
        # if the param is a nested param class
        elif isinstance(param, Params):
            _params_to_omit = _get_omitted_params(param)
            _defaults_to_overwrite = param.__dict__
            flattened_nested_params = _flatten_cls_params(
                type(param),
                use_prefix=use_prefix,
                params_to_omit=_params_to_omit,
                defaults_to_overwrite=_defaults_to_overwrite)
            for n, p in flattened_nested_params.items():
                if n in already_flat_params and not use_prefix:
                    raise KeyError(f'Unable to flatten {cls.__name__} - conflict with param: {n}')
                if not use_prefix:
                    already_flat_params[n] = p
                else:
                    prefix = _create_param_name_prefix(name,
                                                       getattr(cls, '__nested_prefixes__', None))
                    name_variant = prefix + n
                    if name_variant in already_flat_params:
                        raise KeyError(
                            f'Unable to flatten {cls.__name__} - conflict with param: {n} - '
                            f'attempted to use param name: {name_variant} but failed.')
                    already_flat_params[name_variant] = p

    return already_flat_params


def to_argparse(cls: type(Params), **kwargs) -> ArgumentParser:
    """
    Convert a Params class or subclass to an argparse argument parser.
    """
    parser = ArgumentParser(description=cls.__doc__, **kwargs)

    # first, we flatten the cls params (means flattening any nested Params classes)
    flattened_params = _flatten_cls_params(cls)

    # merge any positional arguments
    params_to_delete, positional_param = _merge_positional_params(list(flattened_params.items()))
    for p in params_to_delete:
        del flattened_params[p]
    if positional_param is not None:
        flattened_params['positionals'] = positional_param

    # actually add the params to the argument parser
    for name, param in flattened_params.items():
        _add_param_to_parser(name, param, parser)

    assert sum(getattr(p, 'nargs', '') not in ['+', '*', '?'] for p in vars(cls).values()
               if getattr(p, 'positional', False)) <= 1, \
        'Behavior is undefined for multiple positional arguments with nargs=+|*|?'
    return parser


def _unflatten_params_cls(cls: type(Params),
                          parsed_params: dict,
                          prefix: Optional[str] = None,
                          params_to_omit: Optional[Set[str]] = None) -> Params:
    """
    Recursively construct a Params class or subclass (handles nested Params classes) using values
    parsed from the command line.

    :param cls: The Params class to create an instance of
    :param parsed_params: The params parsed from the command line
    :param prefix: The prefix for all param names
    :param params_to_omit: Params to set to None when de-serializing the parsed params dictionary
    :return: An instance of cls with params set from the parsed params dictionary
    """
    cls_specific_params = {}
    for k, v in cls.__dict__.items():
        if k.startswith('_') or getattr(v, 'hide', False):
            continue
        elif isinstance(v, BaseDescriptor):
            if params_to_omit is not None and k in params_to_omit:
                cls_specific_params[k] = None
                continue
            param_name = prefix + k if k not in parsed_params and prefix is not None else k
            if getattr(v, 'expand', False):
                unexpanded_param = _extract_expanded_param(parsed_params, param_name, v, prefix)
                cls_specific_params.update({k: unexpanded_param})
            else:
                cls_specific_params[k] = parsed_params[param_name]
        elif isinstance(v, Params):
            _params_to_omit = _get_omitted_params(v)
            next_prefix = _create_param_name_prefix(k, getattr(cls, '__nested_prefixes__', None))
            _prefix = prefix + next_prefix if prefix is not None else next_prefix
            cls_specific_params[k] = _unflatten_params_cls(type(v), parsed_params,
                                                           prefix=_prefix,
                                                           params_to_omit=_params_to_omit)
    return cls(**cls_specific_params)


def from_parsed_args(*cls_list, params_namespace: Namespace) -> Tuple[Params]:
    """
    Convert a list of params classes and an argparse parsed namespace to a list of class instances
    """
    params = vars(params_namespace)
    flattened_classes = [_flatten_cls_params(cls) for cls in cls_list]
    # handle positional arguments
    positional_params = {}
    for flattened_cls in flattened_classes:
        positional_params.update({k: v for k, v in flattened_cls.items()
                                  if getattr(v, 'positional', False)})
    if 'positionals' in params:
        parsed_positionals = params['positionals']
        # now we must match the positional params to the list of positional arguments we got back
        matched_pos_args = _parse_positional_arguments(parsed_positionals,
                                                       positional_params)
        params.update(matched_pos_args)

    # actually construct the params classes
    return tuple(_unflatten_params_cls(cls, params) for cls in cls_list)


def create_parser_and_parse_args(*cls_list,
                                 throw_on_unknown: bool = False,
                                 **kwargs) -> Union[Tuple[Params], Params]:
    """
    Outside interface for creating a parser from multiple Params classes and parsing arguments
    """
    parser = to_argparse(merge_param_classes(*cls_list), **kwargs)
    args, argv = parser.parse_known_args()
    if argv != [] and throw_on_unknown:
        raise ValueError(f'Unknown arguments: {argv}')
    return from_parsed_args(*cls_list, params_namespace=args) \
        if len(cls_list) > 1 else from_parsed_args(*cls_list, params_namespace=args)[0]



from typing import Optional

###################
# Unit Conversion #
###################


UNIT_CONVERSION_TABLE = {"GHz": 1.0e9,
                         "MHz": 1.0e6,
                         "ns": 1.0e-9,
                         "μs": 1.0e-6,
                         "us": 1.0e-6,
                         "mV": 1.0e-3,
                         "mW": 1.0e-3,
                         "dBm": 1,
                         "V": 1,
                         "Hz": 1,
                         "s": 1,
                         "Scalar": 1}


def convert_to_si_units(value, unit: Optional[str] = None):
    """
    Convert value to SI units from specified unit.
    """
    if unit is None:
        return value
    try:
        return value * UNIT_CONVERSION_TABLE[unit]
    except KeyError:
        raise KeyError(f'Unit "{unit}" not in conversion table!')
    except TypeError:
        raise TypeError(f'Unit "{unit}" not compatible with value {value} of type {type(value)}')


def unconvert_si_units(value, unit: Optional[str] = None):
    """
    Convert from si units back to "unit" units
    """
    if value is None or unit is None:
        return value
    try:
        return value / UNIT_CONVERSION_TABLE[unit]
    except KeyError:
        raise KeyError(f'Unit "{unit}" not in conversion table!')
    except TypeError:
        raise TypeError(f'Unit "{unit}" not compatible with value {value} of type {type(value)}')



from enum import Enum

import numpy as np
import pytest

from paranormal.parameter_interface import BoolParam, Params, FloatParam, IntParam, StringParam, ListParam, SetParam, \
    EnumParam, GeomspaceParam, ArangeParam, SpanArangeParam
from paranormal.params import SpanLinspaceParam


def test_bool_param():
    class MyParams(Params):
        param = BoolParam(help="A bool param", default=True)

    p = MyParams()

    assert p.param is True
    p.param = False
    assert p.param is False


def test_float_param():
    class MyParams(Params):
        param1 = FloatParam(help="A float param with no default and not required")

        param2 = FloatParam(help="A float param with units", default=4.0, unit="MHz")

    p = MyParams()

    assert p.param1 is None
    p.param1 = 3.0
    assert p.param1 == 3.0

    assert p.param2 == 4.0e6
    p.param2 = 5.0
    assert p.param2 == 5.0e6


def test_int_param():
    class MyParams(Params):
        param1 = IntParam(help="An int param that's required", required=True)

    with pytest.raises(KeyError):
        MyParams()

    p = MyParams(param1=4)
    assert p.param1 == 4


def test_string_param():
    class MyParams(Params):
        param1 = StringParam(help="A string param with a default", default="abc")

    p = MyParams()
    assert p.param1 == "abc"


def test_list_param():
    class MyParams(Params):
        param1 = ListParam(help="A list of ints", subtype=int, default=[1, 2, 3])

        param2 = ListParam(help="A list of floats with units", subtype=float, unit="MHz")

    p1 = MyParams()
    p2 = MyParams()

    assert p1.param1 == p2.param1 == [1, 2, 3]
    p1.param1 = [4, 5, 6]
    assert p1.param1 == [4, 5, 6]
    assert p2.param1 == [1, 2, 3]

    assert p1.param2 is None
    p1.param2 = [1.0, 2.0, 3.0]
    assert p1.param2 == [1.0e6, 2.0e6, 3.0e6]


def test_set_param():
    class MyParams(Params):
        param1 = SetParam(help="A set param")

    p = MyParams(param1={1, 2, 3, 3})

    assert p.param1 == {1, 2, 3}


def test_enum_param():
    class MyEnum(Enum):
        RED = 0
        GREEN = 1
        BLUE = 2

    class MyParams(Params):
        param1 = EnumParam(help="An enum param", cls=MyEnum, default=MyEnum.RED)

    p = MyParams()
    assert p.param1 == MyEnum.RED
    p.param1 = MyEnum.BLUE
    assert p.param1 == MyEnum.BLUE


def test_geomspace_param():
    class MyParams(Params):
        param1 = GeomspaceParam(help="A geometrically spaced list")

    p = MyParams(param1=[1, 2, 3])
    assert np.all(p.param1 == np.geomspace(1, 2, 3))


def test_arange_param():
    class MyParams(Params):
        param1 = ArangeParam(help="An evenly spaced list")

    p = MyParams(param1=[1, 2, 0.1])
    assert len(p.param1) == 10


def test_span_arange_param():
    class MyParams(Params):
        param1 = SpanArangeParam(help="An evenly spaced list created with center, width, step")

    p = MyParams(param1=[1, 2, 0.1])
    assert len(p.param1) == 20


def test_span_linspace_param():
    class MyParams(Params):
        param1 = SpanLinspaceParam(help="An evenly spaced list created with center, width, num")

    p = MyParams(param1=[1, 2, 10])
    assert np.all(p.param1 == np.linspace(0, 2, 10))



from abc import ABC
from enum import Enum, EnumMeta
from typing import Iterable, List, Optional, Set, Tuple
import warnings

import numpy as np

from paranormal.units import convert_to_si_units


__all__ = ['BaseDescriptor', 'BoolParam', 'FloatParam', 'IntParam', 'StringParam', 'ListParam',
           'SetParam', 'ArangeParam', 'EnumParam', 'GeomspaceParam', 'SpanArangeParam',
           'LinspaceParam', 'SpanLinspaceParam']


#################################
# All the Different Param Types #
#################################

class BaseDescriptor(ABC):
    """
    Abstract base type for overriding python descriptor behavior
    """
    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)

    def __get__(self, instance, owner):
        return instance.__dict__.get(self.name, None)

    def __set__(self, instance, value):
        instance.__dict__[self.name] = value

    def __set_name__(self, owner, name):
        self.name = name

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def to_json(self, instance, include_default: bool = True):
        """
        Method to convert this parameter's value to something that's JSON serializable
        """
        return instance.__dict__.get(self.name,
                                     getattr(self, 'default', None) if include_default else None)


class BoolParam(BaseDescriptor):
    """
    A Boolean that overrides the python descriptor behavior
    """
    def __init__(self, *, help: str, default: bool, hide: bool = False, **kwargs):
        """
        :param help: A string that describes what this boolean is for
        :param default: The default value for this boolean
        :param hide: Whether or not to hide this parameter from the argument parser
        :param kwargs: Extra kwargs to modify behavior (not used)
        """
        if 'required' in kwargs or default is None:
            raise ValueError('BoolParam cannot be a required argument '
                             'and must have a default value')
        self.default = default
        self.help = help
        self.hide = hide
        super(BoolParam, self).__init__(**kwargs)

    def __get__(self, instance, owner) -> bool:
        return instance.__dict__.get(self.name, self.default)


class FloatParam(BaseDescriptor):
    """
    A Float that overrides the python descriptor behavior
    """
    def __init__(self, *,
                 help: str,
                 default: Optional[float] = None,
                 hide: bool = False,
                 required: Optional[bool] = None,
                 unit: Optional[str] = None,
                 **kwargs):
        """
        :param help: A string that describes what this float is for
        :param default: The default value for this parameter
        :param hide: Hide this parameter from the argument parser
        :param required: Whether or not this parameter is required when parsing using an
            argument parser
        :param unit: A unit for this parameter (check paranormal.units for possible units)
        :param kwargs: Extra kwargs to modify behavior:
            'choices': list or tuple of all possible choices for this parameter's value - only
                relevant for the argument parser (consider an EnumParam for something more strict)
            'positional' Whether or not to parse this parameter as a positional
        """
        if default is not None and required:
            raise ValueError('Default cannot be specified if required is True!')
        self.default = default
        self.hide = hide
        self.required = required
        self.help = help
        self.unit = unit
        super(FloatParam, self).__init__(**kwargs)

    def __get__(self, instance, owner) -> Optional[float]:
        if self.required and self.name not in instance.__dict__:
            raise ValueError(f'{self.name} is a required argument and must be set first!')
        if instance.__dict__.get(self.name, self.default) is None:
            return None
        return convert_to_si_units(instance.__dict__.get(self.name, self.default), self.unit)


class IntParam(BaseDescriptor):
    """
    An Int that overrides the python descriptor behavior
    """
    def __init__(self, *,
                 help: str,
                 default: Optional[int] = None,
                 hide: bool = False,
                 required: Optional[bool] = None,
                 unit: Optional[str] = None,
                 **kwargs):
        """
        :param help: A string that describes what this int is for
        :param default: The default value for this parameter
        :param hide: Hide this parameter from the argument parser
        :param required: Whether or not this parameter is required when parsing using an
            argument parser
        :param unit: A unit for this parameter (check paranormal.units for possible units)
        :param kwargs: Extra kwargs to modify behavior:
            'choices': list or tuple of all possible choices for this parameter's value - only
                relevant for the argument parser (consider an EnumParam for something more strict)
            'positional' Whether or not to parse this parameter as a positional
        """
        if default is not None and required:
            raise ValueError('Default cannot be specified if required is True!')
        self.default = default
        self.hide = hide
        self.required = required
        self.help = help
        self.unit = unit
        super(IntParam, self).__init__(**kwargs)

    def __get__(self, instance, owner) -> Optional[int]:
        if self.required and self.name not in instance.__dict__:
            raise ValueError(f'{self.name} is a required argument and must be set first!')
        if instance.__dict__.get(self.name, self.default) is None:
            return None
        return convert_to_si_units(instance.__dict__.get(self.name, self.default), self.unit)


class StringParam(BaseDescriptor):
    """
    A string that overrides the python descriptor behavior
    """
    def __init__(self, *,
                 help: str,
                 default: Optional[str] = None,
                 hide: bool = False,
                 required: Optional[bool] = None,
                 **kwargs):
        """
        :param help: A string that describes what this int is for
        :param default: The default value for this parameter
        :param hide: Hide this parameter from the argument parser
        :param required: Whether or not this parameter is required when parsing using an
            argument parser
        :param unit: A unit for this parameter (check paranormal.units for possible units)
        :param kwargs: Extra kwargs to modify behavior:
            'choices': list or tuple of all possible choices for this parameter's value - only
                relevant for the argument parser (consider an EnumParam for something more strict)
            'positional' Whether or not to parse this parameter as a positional
        """
        if default is not None and required:
            raise ValueError('Default cannot be specified if required is True!')
        self.default = default
        self.hide = hide
        self.required = required
        self.help = help
        super(StringParam, self).__init__(**kwargs)

    def __get__(self, instance, owner) -> Optional[str]:
        if self.required and self.name not in instance.__dict__:
            raise ValueError(f'{self.name} is a required argument and must be set first!')
        if instance.__dict__.get(self.name, self.default) is None:
            return None
        return str(instance.__dict__.get(self.name, self.default))


class ListParam(BaseDescriptor):
    """
    A list that overrides the python descriptor behavior
    """
    def __init__(self, *,
                 help: str,
                 subtype: Optional[type] = None,
                 default: Optional[List] = None,
                 hide: bool = False,
                 required: Optional[bool] = None,
                 unit: Optional[str] = None,
                 **kwargs):
        """
        :param help: A string that describes what this list is for
        :param subtype: The type of objects that make up this list (specify if you want to parse
            using an argument parser)
        :param default: The default value for this parameter
        :param hide: Hide this parameter from the argument parser
        :param required: Whether or not this parameter is required when parsing using an
            argument parser
        :param unit: A unit for this parameter (check paranormal.units for possible units)
        :param kwargs: Extra kwargs to modify behavior:
            'choices': list or tuple of all possible choices for this items in the list - only
                relevant for the argument parser (consider an EnumParam for something more strict)
            'positional' Whether or not to parse this parameter as a positional
            'nargs': How many arguments this parameter should take when parsing from the command
                line
        """
        if default is not None and required:
            raise ValueError('Default cannot be specified if required is True!')
        self.default = default
        self.hide = hide
        self.required = required
        self.help = help
        self.unit = unit
        self.subtype = subtype
        super(ListParam, self).__init__(**kwargs)

    def __get__(self, instance, owner) -> Optional[List]:
        if self.required and self.name not in instance.__dict__:
            raise ValueError(f'{self.name} is a required argument and must be set first!')
        if instance.__dict__.get(self.name, self.default) is None:
            return None
        return list(convert_to_si_units(v, self.unit)
                    for v in instance.__dict__.get(self.name, self.default))


class SetParam(BaseDescriptor):
    """
    A set that overrides python descriptor behavior
    """
    def __init__(self, *,
                 help: str,
                 subtype: type = str,
                 default: Optional[Set] = None,
                 hide: bool = False,
                 required: Optional[bool] = None,
                 unit: Optional[str] = None,
                 **kwargs):
        """
        :param help: A string that describes what this set is for
        :param subtype: The type of objects that make up this list (specify if you want to parse
            using an argument parser)
        :param default: The default value for this parameter
        :param hide: Hide this parameter from the argument parser
        :param required: Whether or not this parameter is required when parsing using an
            argument parser
        :param unit: A unit for this parameter (check paranormal.units for possible units)
        :param kwargs: Extra kwargs to modify behavior:
            'choices': list or tuple of all possible choices for this items in the list - only
                relevant for the argument parser (consider an EnumParam for something more strict)
            'positional' Whether or not to parse this parameter as a positional
            'nargs': How many arguments this parameter should take when parsing from the command
                line
        """
        if default is not None and required:
            raise ValueError('Default cannot be specified if required is True!')
        self.default = default
        self.hide = hide
        self.required = required
        self.help = help
        self.unit = unit
        self.subtype = subtype
        super(SetParam, self).__init__(**kwargs)

    def __get__(self, instance, owner) -> Optional[Set]:
        if self.required and self.name not in instance.__dict__:
            raise ValueError(f'{self.name} is a required argument and must be set first!')
        if instance.__dict__.get(self.name, self.default) is None:
            return None
        return set(convert_to_si_units(v, self.unit)
                   for v in instance.__dict__.get(self.name, self.default))


class EnumParam(BaseDescriptor):
    """
    An enum that overrides python descriptor behavior
    """
    def __init__(self, *,
                 help: str,
                 cls: EnumMeta,
                 default: Optional[Enum] = None,
                 hide: bool = False,
                 required: Optional[bool] = None,
                 **kwargs):
        """
        :param help: A string that describes what this enum is for
        :param cls: The enum class this parameter represents
        :param default: The default value for this parameter
        :param hide: Hide this parameter from the argument parser
        :param required: Whether or not this parameter is required when parsing using an
            argument parser
        :param kwargs: Extra kwargs to modify behavior:
            'positional' Whether or not to parse this parameter as a positional
        """
        if default is not None and required:
            raise ValueError('Default cannot be specified if required is True!')
        if default not in cls and default is not None:
            raise ValueError(f'Default {default} is not a member of {cls}')
        self.cls = cls
        self.default = default
        self.hide = hide
        self.required = required
        self.help = help
        super(EnumParam, self).__init__(**kwargs)

    def __get__(self, instance, owner) -> Optional[Enum]:
        if self.required and self.name not in instance.__dict__:
            raise ValueError(f'{self.name} is a required argument and must be set first!')
        return instance.__dict__.get(self.name, self.default)

    def __set__(self, instance, value):
        if value is None or value in self.cls:
            instance.__dict__[self.name] = value
        elif value in self.cls.__members__:
            instance.__dict__[self.name] = self.cls[value]
        else:
            raise KeyError(f'{value} is not a valid member of {self.cls.__name__}')

    def to_json(self, instance, include_default: bool = True):
        default = getattr(self, 'default', None) if include_default else None
        tmp = instance.__dict__.get(self.name, default)
        return tmp.name if tmp is not None else None


###############################
# Numpy function based params #
###############################


def _check_numpy_fn_param_value(name: str, value: Iterable, nargs: int) -> bool:
    """
    Check whether a value is fit for use in a three argument numpy function
    """
    if (isinstance(value, (tuple, list)) and len(value) == nargs and
            not any([i is None for i in value])):
        return True
    warnings.warn(f'Param {name} does not match format required (List or Tuple of length '
                  f'{nargs} without any None values. Default behavior will not work.')
    return False


class NumpyFunctionParam(BaseDescriptor):
    """
    A parameter that is set with 3 arguments and returns the result of a numpy function called
    on those arguments if those arguments satisfy _check_numpy_fn_param_value
    """
    nargs = 3

    def __init__(self, *,
                 help: str,
                 default: Optional[Tuple] = None,
                 hide: bool = False,
                 required: Optional[bool] = None,
                 unit: Optional[str] = None,
                 **kwargs):
        """
        :param help: A string that describes what this parameter is for
        :param subtype: The type of objects that make up this list (specify if you want to parse
            using an argument parser)
        :param default: The default value for this parameter
        :param hide: Hide this parameter from the argument parser
        :param required: Whether or not this parameter is required when parsing using an
            argument parser
        :param unit: A unit for this parameter (check paranormal.units for possible units)
        :param kwargs: Extra kwargs to modify behavior:
            'choices': list or tuple of all possible choices for this parameter's value - only
                relevant for the argument parser (consider an EnumParam for something more strict)
            'expand': Whether or not to expand this parameter into sub arguments when argument
                parsing
            'positional' Whether or not to parse this parameter as a positional
            'nargs': How many arguments this parameter should take when parsing from the command
                line - overwrites the default of 3
        """
        if default is not None and required:
            raise ValueError('Default cannot be specified if required is True!')
        if not isinstance(default, tuple) and default is not None:
            raise ValueError('Default must be either a tuple or None')
        self.default = default
        self.hide = hide
        self.required = required
        self.help = help
        self.unit = unit
        super(NumpyFunctionParam, self).__init__(**kwargs)

    def __get__(self, instance, owner) -> Optional[np.ndarray]:
        if self.required and self.name not in instance.__dict__:
            raise ValueError(f'{self.name} is a required argument and must be set first!')
        v = instance.__dict__.get(self.name, self.default)
        if not _check_numpy_fn_param_value(self.name, v, self.nargs):
            return v
        return convert_to_si_units(self._numpy_function(v), self.unit)

    def __set__(self, instance, value):
        _check_numpy_fn_param_value(self.name, value, self.nargs)
        instance.__dict__[self.name] = value

    def _numpy_function(self, value: Iterable) -> np.ndarray:
        raise NotImplementedError()

    def to_json(self, instance, include_default: bool = True):
        """
        Convert this parameter to a json serializable value
        """
        v = instance.__dict__.get(self.name, self.default)
        if _check_numpy_fn_param_value(self.name, v, self.nargs):
            return super(NumpyFunctionParam, self).to_json(instance, include_default)
        elif isinstance(v, np.ndarray):
            return v.tolist()
        else:
            return v


class GeomspaceParam(NumpyFunctionParam):
    """
    A parameter that is set with 3 arguments (start, stop, num) and returns the geomspace function
    called on those arguments if those arguments satisfy _check_numpy_fn_param_value
    """
    def __set__(self, instance, value):
        if _check_numpy_fn_param_value(self.name, value, self.nargs):
            assert value[0] != 0
        super(GeomspaceParam, self).__set__(instance, value)

    def _numpy_function(self, value: Iterable):
        return np.geomspace(*value)


class ArangeParam(NumpyFunctionParam):
    """
    A parameter that is set with 3 arguments (start, stop, step) and returns the arange function
    called on those arguments if those arguments satisfy _check_numpy_fn_param_value
    """
    def _numpy_function(self, value: Iterable):
        return np.arange(*value)


class SpanArangeParam(NumpyFunctionParam):
    """
    A parameter that is set with 3 arguments (center, width, step) and returns the arange function
    called on (center - 0.5 * width, center + 0.5 * width, step) if those arguments satisfy
    _check_numpy_fn_param_value
    """
    def _numpy_function(self, value: Iterable):
        center, width, step = value[0], value[1], value[2]
        return np.arange(center - 0.5 * width, center + 0.5 * width, step)


class LinspaceParam(NumpyFunctionParam):
    """
    A parameter that is set with 3 arguments (start, stop, num) and returns the linspace function
    called on those arguments if those arguments satisfy _check_numpy_fn_param_value
    """
    def _numpy_function(self, value: Iterable):
        return np.linspace(value[0], value[1], int(value[2]))


class SpanLinspaceParam(NumpyFunctionParam):
    """
    A parameter that is set with 3 arguments (center, width, num) and returns the linspace function
    called on (center - 0.5 * width, center + 0.5 * width, num) if those arguments satisfy
    _check_numpy_fn_param_value
    """
    def _numpy_function(self, value: Iterable):
        center, width, num = value[0], value[1], value[2]
        return np.linspace(center - 0.5 * width, center + 0.5 * width, int(num))



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['paranormal']

package_data = \
{'': ['*'], 'paranormal': ['test/*']}

install_requires = \
['PyYAML>=5.1,<6.0', 'numpy>=1.16,<2.0', 'pampy>=0.2.1,<0.3.0']

setup_kwargs = {
    'name': 'paranormal',
    'version': '0.1.8',
    'description': 'Coherent management of large parameter lists in Python',
    'long_description': None,
    'author': 'Schuyler Fried',
    'author_email': 'schuylerfried@gmail.com',
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>3.6',
}


setup(**setup_kwargs)



from argparse import Namespace
from enum import Enum
import json
import mock
import pytest
import tempfile

import numpy as np

from paranormal.parameter_interface import *
from paranormal.params import *


################################
# Helper Classes and Functions #
################################


class E(Enum):
    X = 0
    Y = 1


class P(Params):
    b = BoolParam(default=True, help='a boolean!')
    i = IntParam(default=1, help='an integer!')
    f = FloatParam(default=0.5, help='A float!', unit='MHz')
    r = IntParam(required=True, help='An integer that is required!')
    e = EnumParam(cls=E, help='an enum', default=E.X)
    l = ListParam(subtype=int, default=[0, 1, 2], help='a list')
    a = ArangeParam(help='arange param', default=(0, 100, 5), unit='us')


class Colors(Enum):
    RED = 0
    BLUE = 1
    GREEN = 2
    YELLOW = 3


class MySummer(Params):
    dpw_s = LinspaceParam(help='Drinks per weekend', expand=True, default=(0, None, 15),
                          prefix='s_')
    c = EnumParam(cls=Colors, default=Colors.BLUE, help='Color of the sky')
    t = FloatParam(help='Time spent sunbathing before I burn', default=60, unit='ns')
    f = FloatParam(help='Frequency of birds chirping', unit='MHz', default=None)
    do_something_crazy = BoolParam(default=False, help='Do something crazy')


class MyWinter(Params):
    s = FloatParam(default=12, help='hours sleeping per 24 hrs')
    hib = BoolParam(default=False, help='Whether or not to hibernate')
    dpw_w = LinspaceParam(help='Drinks per weekend', expand=True, default=(0, None, 15),
                          prefix='w_')


class MySpring(Params):
    flowers = FloatParam(default=12, help='flowers sprouting per day')
    dpw_s = LinspaceParam(help='Drinks per weekend', expand=True, default=(0, None, 15),
                          prefix='sp_')


class PositionalsA(Params):
    x = FloatParam(positional=True, help='a positional float')
    y = StringParam(positional=True, help='a positional string')
    z = LinspaceParam(positional=True, help='a positional linspace')


class PositionalsB(Params):
    a = IntParam(positional=True, help='a positional int')
    b = IntParam(help='an int')


class FreqSweep(Params):
    freqs = LinspaceParam(help='freqs', expand=True, default=(10, 20, 30), unit='MHz',
                          prefix='f_')
    times = LinspaceParam(help='times', expand=True, default=(100, 200, 50), unit='us',
                          prefix='t_')


class TimeSweep(Params):
    times = LinspaceParam(help='times', expand=True, default=(100, 500, 20), unit='ns',
                          prefix='t_')


class DoubleSweep(Params):
    freq_sweep = FreqSweep()
    time_sweep = TimeSweep()


class HiddenWinterSchedule(Params):
    winter = MyWinter(s='__hide__', dpw_w='__hide__')


class HiddenFall(Params):
    leaf = FloatParam(default=2, help='number of falling leaves caught', hide=True)
    pumpkin = BoolParam(default=True, help='Whether the pumpkin got carved', hide=False)


def _compare_two_param_item_lists(a, b):
    for (k, v), (k_cor, v_cor) in zip(a, b):
        assert k == k_cor
        if isinstance(v, np.ndarray):
            assert np.allclose(v, v_cor)
        else:
            assert v == v_cor


#####################
# Actual Unit Tests #
#####################


def test_params():

    p = P(b=False, i=2, f=0.2, r=5)
    correct_values = [('b', False), ('i', 2), ('f', 0.2e6), ('r', 5), ('e', E.X), ('l', [0, 1, 2]),
                      ('a', np.array([0,  5, 10, 15, 20, 25, 30, 35, 40, 45, 50, 55, 60, 65, 70,
                                      75, 80, 85, 90, 95]) * 1e-6)]
    _compare_two_param_item_lists(p.items(), correct_values)

    # test with an argument that isn't in P
    with pytest.raises(KeyError):
        P(x=5, r=2)

    # Try without providing r
    with pytest.raises(KeyError):
        P()

    # test si_set
    p.si_set('f', 0.3e6)
    assert p.f == 0.3e6
    p.si_set('a', [1.2e-6, 1.5e-6, 100])
    assert np.allclose(p.a, np.arange(1.2, 1.5, 100) * 1e-6)


def test_json_serialization():

    # to_json
    p = P(r=5)
    s = json.dumps(to_json_serializable_dict(p, include_defaults=False))
    assert s == '{"r": 5, "_type": "P", "_module": "test_parameter_interface"}'

    s = json.dumps(to_json_serializable_dict(p, include_defaults=True))
    assert s == '{"b": true, "i": 1, "f": 0.5, "r": 5, "e": "X", "l": [0, 1, 2], ' \
                '"a": [0, 100, 5], "_type": "P", "_module": "test_parameter_interface"}'
    # from json
    p = P(r=3)
    j = json.loads(json.dumps(to_json_serializable_dict(p, include_defaults=True)))
    assert p == from_json_serializable_dict(j)

    j = json.loads(json.dumps(to_json_serializable_dict(p, include_defaults=False)))
    assert p == from_json_serializable_dict(j)

    # test with hiding some nested params
    y = HiddenWinterSchedule()
    j = json.loads(json.dumps(to_json_serializable_dict(y)))
    assert y == from_json_serializable_dict(j)

    # test with setting nested params
    y = DoubleSweep()
    y.freq_sweep.freqs = (100, 200, 300)
    s = json.dumps(to_json_serializable_dict(y))
    assert s == '{"freq_sweep": {"freqs": [100, 200, 300], "times": [100, 200, 50], ' \
                '"_type": "FreqSweep", "_module": "test_parameter_interface"}, ' \
                '"time_sweep": {"times": [100, 500, 20], "_type": "TimeSweep", ' \
                '"_module": "test_parameter_interface"}, "_type": "DoubleSweep", ' \
                '"_module": "test_parameter_interface"}'

    # test changing defaults of nested params in one class instance doesn't change the default
    # for all class instances
    x = DoubleSweep()
    assert x != y

    # ensure warnings are working when setting something to __hide__
    y.freq_sweep.freqs = '__hide__'
    with pytest.warns(UserWarning):
        to_json_serializable_dict(y)


def test_yaml_serialization():
    p = P(r=10)
    temp = tempfile.NamedTemporaryFile(delete=False, mode='w+t')
    with mock.patch('paranormal.parameter_interface.open') as o:
        o.return_value = temp
        to_yaml_file(p, 'mock.yaml', include_defaults=True)
    with open(temp.name) as yaml_file:
        assert yaml_file.read() == 'b: true\ni: 1\nf: 0.5\nr: 10\ne: X\nl:\n- 0\n- 1\n- 2\na: ' \
                                   '!!python/tuple\n- ' \
                                   '0\n- 100\n- 5\n_type: P\n_module: test_parameter_interface\n'

    # test yaml dumping with alphabetical reorder
    p = P(r=10)
    temp = tempfile.NamedTemporaryFile(delete=False, mode='w+t')
    with mock.patch('paranormal.parameter_interface.open') as o:
        o.return_value = temp
        to_yaml_file(p, 'mock.yaml', include_defaults=True, sort_keys=True)
    with open(temp.name) as yaml_file:
        assert yaml_file.read() == '_module: test_parameter_interface\n_type: P\na: ' \
                                   '!!python/tuple\n- 0\n- 100' \
                                   '\n- 5\nb: true\ne: X\nf: 0.5\ni: 1\nl:\n- 0\n- 1\n- 2\nr: 10\n'


def test_merge_params():

    merged_positionals = merge_param_classes(PositionalsA, PositionalsB)
    for attr in PositionalsA():
        if getattr(PositionalsA.__dict__.get(attr), 'positional', False):
            assert merged_positionals.__dict__.get(attr, None) is None
        else:
            assert merged_positionals.__dict__.get(attr, None) == PositionalsA.__dict__.get(attr)
    for attr in PositionalsB():
        if getattr(PositionalsB.__dict__.get(attr), 'positional', False):
            assert merged_positionals.__dict__.get(attr, None) is None
        else:
            assert merged_positionals.__dict__.get(attr, None) == PositionalsB.__dict__.get(attr)

    # try without merging the positionals
    merged_positionals = merge_param_classes(PositionalsA, PositionalsB,
                                             merge_positional_params=False)
    for attr in PositionalsA():
        assert merged_positionals.__dict__.get(attr, None) == PositionalsA.__dict__.get(attr)
    for attr in PositionalsB():
        assert merged_positionals.__dict__.get(attr, None) == PositionalsB.__dict__.get(attr)

    # try with conflicting parameters
    with pytest.raises(ValueError):
        merge_param_classes(FreqSweep, TimeSweep)

    # try with nested classes
    class YearlySchedule(Params):
        winter = MyWinter()
        summer = MySummer()
        x = IntParam(help='an int')
    merged = merge_param_classes(DoubleSweep, YearlySchedule,
                                 merge_positional_params=False)
    for attr in YearlySchedule():
        assert merged.__dict__.get(attr, None) == YearlySchedule.__dict__.get(attr)
    for attr in DoubleSweep():
        assert merged.__dict__.get(attr, None) == DoubleSweep.__dict__.get(attr)


def test_append_params_attributes():
    class A(Params):
        i = IntParam(help='int', default=1)

    class B(Params):
        f = FloatParam(help='float', default=2.0)
        z = IntParam(help='another int')

    append_params_attributes(A, B, do_not_copy=['z'], override_dictionary={
        'f': {'help': 'a better float', 'default': 3.0, 'positional': True}})
    a = A()
    assert a.f == 3.0
    assert A.__dict__['f'].help == 'a better float'
    assert A.__dict__['f'].positional

    with pytest.raises(AttributeError):
        assert a.z


def test_to_argparse():

    parser = to_argparse(MySummer)

    # parse with the defaults
    args = parser.parse_args([])
    # dpw was expanded and will still be in the namespace, just as None
    assert args == Namespace(c=Colors.BLUE, do_something_crazy=False, t=60, dpw_s=None,
                             s_start=0, s_stop=None, s_num=15, f=None)

    args = parser.parse_args(
        '--f 120 --c GREEN --s_start 20 --s_stop 600 --s_num 51 --do_something_crazy'.split(' '))
    assert args == Namespace(c='GREEN', do_something_crazy=True, t=60, dpw_s=None,
                             s_start=20, s_stop=600, s_num=51, f=120)

    # try with an argument that isn't part of the parser
    with pytest.raises(SystemExit):
        parser.parse_args(
            '--do_soooomething_crazy --f 120'.split(' '))


    class YearlySchedule(Params):
        winter = MyWinter()
        summer = MySummer(f=360)

    parser = to_argparse(YearlySchedule)
    args = parser.parse_args(
        '--summer_c RED --winter_s 22 --summer_s_start 20 --summer_s_stop 600 --winter_w_start 20 '
        '--winter_w_stop 200 --winter_hib'.split(' '))
    assert args == Namespace(summer_c='RED', summer_do_something_crazy=False, summer_dpw_s=None,
                             summer_f=360, summer_s_num=15, summer_s_start=20.0,
                             summer_s_stop=600.0, summer_t=60, winter_dpw_w=None, winter_hib=True,
                             winter_s=22.0, winter_w_num=15, winter_w_start=20.0,
                             winter_w_stop=200.0)

    class YearlySchedule(Params):
        winter = MyWinter()
        summer = MySummer()
        spring = MySpring()

    parser = to_argparse(YearlySchedule)
    args = parser.parse_args([])
    assert args == Namespace(summer_c=Colors.BLUE, summer_do_something_crazy=False,
                             summer_dpw_s=None, summer_f=None, summer_s_num=15, summer_s_start=0,
                             summer_s_stop=None, summer_t=60, winter_dpw_w=None, winter_hib=False,
                             winter_s=12, winter_w_num=15, winter_w_start=0, winter_w_stop=None,
                             spring_dpw_s=None, spring_flowers=12, spring_sp_num=15,
                             spring_sp_start=0, spring_sp_stop=None)


    # Make sure conflicting params are resolved
    parser = to_argparse(DoubleSweep)
    args = parser.parse_args([])
    assert args == Namespace(freq_sweep_f_num=30, freq_sweep_f_start=10, freq_sweep_f_stop=20,
                             freq_sweep_freqs=None, freq_sweep_t_num=50, freq_sweep_t_start=100,
                             freq_sweep_t_stop=200, freq_sweep_times=None, time_sweep_t_num=20,
                             time_sweep_t_start=100, time_sweep_t_stop=500, time_sweep_times=None)

    # make sure check that requires prefixes if expand=True for multiple classes is working
    class BadFreqSweep(Params):
        freqs = LinspaceParam(help='freqs', expand=True)
        times = LinspaceParam(help='times', expand=True)

    with pytest.raises(ValueError):
        to_argparse(BadFreqSweep)

    # test with hiding some nested params
    parser = to_argparse(HiddenWinterSchedule)
    args = parser.parse_args([])
    assert vars(args) == {'winter_hib': False}

    # test by overriding prefix behavior
    class PrefixMania(Params):
        x = ArangeParam(help='some arange', default=(0, 10, 20))

    class NestedPrefixMania(Params):
        a = PrefixMania()
        b = PrefixMania()
        __nested_prefixes__ = {'a': 'ayy', 'b': None}

    parser = to_argparse(NestedPrefixMania)
    args = parser.parse_args([])
    assert vars(args) == {'ayy_x': (0, 10, 20), 'x': (0, 10, 20)}


def test_from_parsed_args():

    parser = to_argparse(MyWinter)
    y = from_parsed_args(MyWinter, params_namespace=parser.parse_args([]))[0]
    correct_items = [('s', 12), ('hib', False), ('dpw_w', [0, None, 15])]
    _compare_two_param_item_lists(y.items(), correct_items)

    args = parser.parse_args('--w_stop 10 --w_num 11 --hib --s 35'.split(' '))
    y = from_parsed_args(MyWinter, params_namespace=args)[0]
    correct_items = [('s', 35), ('hib', True), ('dpw_w', np.arange(0, 11))]
    _compare_two_param_item_lists(y.items(), correct_items)

    class YearlySchedule(Params):
        winter = MyWinter()
        summer = MySummer(f=None)

    # test that nested classes work
    parser = to_argparse(YearlySchedule)
    args = parser.parse_args(
        '--summer_c RED --winter_s 22 --summer_s_start 20 --summer_s_stop 600 --winter_w_start 20 '
        '--winter_w_stop 200 --winter_hib'.split(' '))
    y = from_parsed_args(YearlySchedule, params_namespace=args)[0]
    correct_items = [('winter', MyWinter(s=22, dpw_w=[20, 200, 15], hib=True)),
                     ('summer', MySummer(c=Colors.RED, dpw_s=[20, 600, 15]))]
    _compare_two_param_item_lists(y.items(), correct_items)

    # test that nested classes with positionals work
    class PositionalsC(Params):
        a_pos = PositionalsA()
        b_pos = PositionalsB()

    parser = to_argparse(PositionalsC)
    args = parser.parse_args(
        '1.0 hey 0.0 1.0 22.0 1 --b 10'.split(' '))
    y = from_parsed_args(PositionalsC, params_namespace=args)[0]
    correct_items = [('a_pos', PositionalsA(x=1.0, z=[0.0, 1.0, 22.0], y='hey')),
                     ('b_pos', PositionalsB(a=1, b=10))]
    _compare_two_param_item_lists(y.items(), correct_items)

    parser = to_argparse(DoubleSweep)

    # make sure if you pass an expanded param, an error is thrown
    with pytest.raises(AssertionError):
        args = parser.parse_args([])
        setattr(args, 'freq_sweep_times', [0, 100, 200])
        from_parsed_args(DoubleSweep, params_namespace=args)

    args = parser.parse_args(
        '--time_sweep_t_start 20 --time_sweep_t_stop 30 --freq_sweep_f_stop 40'.split(' '))
    y = from_parsed_args(DoubleSweep, params_namespace=args)[0]
    correct_items = [('freq_sweep', FreqSweep(freqs=[10, 40.0, 30])),
                     ('time_sweep', TimeSweep(times=[20, 30, 20]))]
    _compare_two_param_item_lists(y.items(), correct_items)

    # test with hiding some nested params
    parser = to_argparse(HiddenWinterSchedule)
    args = parser.parse_args([])
    y = from_parsed_args(HiddenWinterSchedule, params_namespace=args)[0]
    correct_items = [('winter', MyWinter(dpw_w=None, s=None))]
    _compare_two_param_item_lists(y.items(), correct_items)

    # test that hide param works
    parser = to_argparse(HiddenFall)
    help = parser.format_help()
    assert 'pumpkin' in help
    assert 'leaf' not in help


def test_create_parser_and_parse_args():
    # test that nested classes with positionals work
    class PositionalsC(Params):
        a_pos = PositionalsA()
        b_pos = PositionalsB()

    correct_items = [('a_pos', PositionalsA(x=1.0, z=[0.0, 1.0, 22.0], y='hey')),
                     ('b_pos', PositionalsB(a=1, b=10))]
    with mock.patch('paranormal.parameter_interface.ArgumentParser.parse_known_args') as pa:
        pa.return_value = (Namespace(b=10, positionals=['1.0', 'hey', '0.0', '1.0', '22.0', '1']),
                           [])
        y = create_parser_and_parse_args(PositionalsC)
    _compare_two_param_item_lists(y.items(), correct_items)

    # test that nested classes work:
    class YearlySchedule(Params):
        winter = MyWinter()
        summer = MySummer(f=None)
        __nested_prefixes__ = {'winter': None, 'summer': None}

    with mock.patch('paranormal.parameter_interface.ArgumentParser.parse_known_args') as pa:
        pa.return_value = (Namespace(c='RED', do_something_crazy=False, dpw_s=None, dpw_w=None,
                                     f=None, hib=True, s=22.0, s_num=15, s_start=20.0, s_stop=600.0,
                                     t=60, w_num=15, w_start=20.0, w_stop=200.0), [])
        y = create_parser_and_parse_args(YearlySchedule)
    correct_items = [('winter', MyWinter(s=22, dpw_w=[20, 200, 15], hib=True)),
                     ('summer', MySummer(c=Colors.RED, dpw_s=[20, 600, 15]))]
    _compare_two_param_item_lists(y.items(), correct_items)

